//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//      PsfPosition.h:       --- Position utilities for PSF class
//
//      DESCRIPTION:
//              An instance of class <PsfPosition> receives a dataset and/or other
//		parameters as input to storing position information of the required
//		PSF. The main purpose of using <PsfPosition> is to get the center
//		of the PSF. It contains the following information:
//                      o PosWCS (from the dataset)
//                      o DETX/DETY, theta/phi of the center
//                      o source coordinates and position of the center
//                      o other necessary instrument and observation information
//
//      IMPLEMENTATION IDEA:
//              <PsfPosition> uses <PosWCS> and <PositionUtils> to calculate coordinate
//		transfermation, <PsfRegion> to get the area. It reads necessary 
//		information from datase.

#ifndef PSFPOSITION_H
#define PSFPOSITION_H

#include "Dal.h"
#include "CalServer.h"
#include "PosWCS.h"
#include "PsfRegion.h"

using namespace std;
class CoverFBox : public virtual FBox {
public:
	CoverFBox() : FBox(), _xOrgDetSize(0), _yOrgDetSize(0), _pa(0), _xAngle(0), _yAngle(0) {}
	CoverFBox(const FPoint &refPoint, float xWidth, float yWidth, float xo, float yo, float pa, float xAngle, float yAngle, bool included = true) :
		FBox(refPoint, xWidth, yWidth, included), _xOrgDetSize(xo), _yOrgDetSize(yo), _pa(pa), _xAngle(xAngle), _yAngle(yAngle) {_refPoint = refPoint;}

	float xAngle() const { return _xAngle; }
	float yAngle() const { return _yAngle; }
	float pa() const { return _pa; }
	float xOrgDetSize() const { return _xOrgDetSize; }
	float yOrgDetSize() const { return _yOrgDetSize; }
private:
	float	_xOrgDetSize;	// original width of image in DET005
	float	_yOrgDetSize;	// original hight of image in DET005
	float	_pa;		// camera rotation angle
	float	_xAngle;	// angle that control horizontal direction flip
	float	_yAngle;	// angle that control vertical direction flip
};

class PsfPosition {
public:
	PsfPosition() : _dataSet(0), _ccdnr(1), _region(0), _posWCS(0) {}
	PsfPosition(const string & dataSetName, const string & expr = "") : _dataSet(0), _region(0), _posWCS(0) { setDataSet(dataSetName, expr); }
	PsfPosition(DataSet * pDataSet, const string & expr = "") : _dataSet(0), _region(0), _posWCS(0) { setDataSet(pDataSet, expr); }

	~PsfPosition() { if (_region) delete _region; if (_posWCS) delete _posWCS; }

	// property accessors
	void setDataSet(DataSet * pDataSet, const string & expr = "", bool init = true);
	void setDataSet(const string & dataSetName, const string & expr = "", bool init = true);

	float theta() const { return _theta; }
	float phi() const { return _phi; }
	float detx() const { return _detx; }
	float dety() const { return _dety; }
	
	const string & instrument() const { return _instrument; }
	const string & dataSet() const { return _dataSetName; }
	const string & expression() const { return _expr; }
	const string & obsStart() const { return _obsStart; }
	const string & obsEnd() const { return _obsEnd; }
	const string & dataMode() const { return _datamode; }

	int ccdnr() const { return _ccdnr; }
	Xmm::Instrument instrumentID() const { return _instrumentID; }
	Xmm::Mode windowMode() const { return _windowMode; }

	const PsfRegion * region() const { return _region; }
	const PosWCS * posWCS() const { return _posWCS;}
	CoverFBox coverBox() const;

	void initialise(const string & expr = "");

private:
	void setRegion(Block * table);
	void setPosition();
	void setCcdnr();
	void readWindowMode();
	double maxDist(double, double, double, double) const;

protected:
	float 		_theta, _phi;
	float 		_detx, _dety;
	
	string 		_dataSetName;
	DataSet * 	_dataSet;
	
	string 		_instrument;
	Xmm::Instrument	_instrumentID;

	int		_ccdnr;
	Xmm::Mode	_windowMode;

	string		_expr;

	string		_obsStart;
	string		_obsEnd;
	string		_datamode;

	PsfRegion *	_region;
	PosWCS *	_posWCS;
};

ostream & operator<< (ostream & out, const CoverFBox & box);
ostream & operator<< (ostream & out, const PsfPosition & pos);
#endif
