//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//      PsfRegion.cc:      --- implementation of <PsfRegion> class
//
//      DESCRIPTION:
//              Implementation of <PsfRegion> class

#include <math.h>
#include <sstream>

// stringutil.h is under /data/share/sas/reltrack/xmmsas_20070615_1801/packages/utils/src/, executable is in libutils
// lowercase() is defined in stringutil.h
#include "stringutil.h"

#include "PsfRegion.h"

void PsfRegion::setCoordType (const string & type) {
	_CoordType = type;
}

void PsfRegion::setExpression(const string & expr) {
	_Expression = expr;
	
	if (expr.empty() && size()) {
		if (_CoordType == "DET")
			_Expression = at(0).toSelectExpression("DETX", "DETY");
		else if (_CoordType == "RAW")
			_Expression = at(0).toSelectExpression("RAWX", "RAWY");
		else 
			_Expression = at(0).toSelectExpression("X", "Y");
	}
}

void PsfRegion::setRegion(const string & expr) {
	_Expression = expr;

	// split expression [i.e. (DETX,DETY) IN circle(132.5,-232.5,1843.58)] into three parts: coord_info , shape_type, and shape_info
	// first, find the coordinate type information, i.e. (DETX,DETY)
	string::size_type pos_start = _Expression.find('(');
	string::size_type pos_end = _Expression.find(')');
	if (pos_start == string::npos || pos_end == string::npos) throw InvalidRegion();

	string coord_info = lowercase(_Expression.substr(pos_start+1, pos_end-pos_start-1));  // coordinate type information
	string::size_type pos_mid = pos_end+1;	// remember the start position for the rest characters

	// second, find the shape type information, i.e. circle
	pos_start = _Expression.rfind('(');
	pos_end = _Expression.rfind(')');
        if (pos_start == string::npos || pos_end == string::npos || pos_mid == string::npos) throw InvalidRegion();

	string shape_type = _Expression.substr(pos_mid, pos_start-pos_mid);

	pos_mid = shape_type.rfind(' ');
	shape_type = lowercase(shape_type.substr(pos_mid));

	// third, find the shape parameter information
	string shape_info = _Expression.substr(pos_start+1, pos_end-1);
	for (string::iterator i = shape_info.begin(); i < shape_info.end(); i++)
		if ( *i == ',' ) *i = ' ';
	
	// set GeoRegion
	FShape * shape = 0;
	float xref, yref, inner[2], outer[2], rot[2];
	istringstream ipara(shape_info);

	ipara >> xref >> yref;		// read in reference point
	FPoint refPoint(xref, yref);	// reference point

	if ( shape_type.find("circle") != string::npos ) {
		// circle, the third parameter is radius
		ipara >> outer[0];
		shape = new FCircle(refPoint, outer[0]);
	}
	else if ( shape_type.find("annulus") != string::npos ) {
		// annulus, the third & fourth parameter are radius of inner & outer circle
		ipara >> inner[0] >> outer[0];
		shape = new FAnnulus(refPoint, inner[0], outer[0]);
	}
	else if ( shape_type.find("elliptannulus") != string::npos) {
		// elliptannulus, the 3rd & 4th & 7th parameters are for inner ellipse, 5th & 6th & 8th parameters  
		// are for outer ellipse
		ipara >> inner[0] >> inner[1] >> outer[0] >> outer[1] >> rot[0] >> rot[1];
		shape = new FElliptAnnulus(refPoint, inner[0], inner[1], outer[0], outer[1], rot[0], rot[1]);
	}
        else if ( shape_type.find("ellipse") != string::npos ) {
                // ellipse, the third & fourth parameters are major & minor radius, fifth paramter is rotation angle
                ipara >> outer[0] >> outer[1] >> rot[0];
                shape = new FEllipse(refPoint, outer[0], outer[1], rot[0]);
        }
 	else if ( shape_type.find("rotbox") != string::npos ) {
		// rotbox, the 3rd & 4th parameters are the width and height, 5th parameter is rotation
                ipara >> outer[0] >> outer[1] >> rot[0];
                shape = new FRotBox(refPoint, outer[0], outer[1], rot[0]);
	}
        else if ( shape_type.find("box") != string::npos ) {
                // box, the 3rd & 4th parameters are the width and height
                ipara >> outer[0] >> outer[1];
                shape = new FBox(refPoint, outer[0], outer[1]);
        }
 	else throw InvalidRegion();

	if (shape) {
		Component comp(1, shape);
		push_back(comp);
	}

	// set the coordinate type
	if ( coord_info.find("det") != string::npos ) 		// DETX/DETY 0.05 coordinate
		_CoordType = "DET";
	else if ( coord_info.find("raw") != string::npos )	// RAWX/RAWY coordinate
		_CoordType = "RAW";
	else _CoordType = "POS";				// X/Y coordinate

	// set cover box
	if ( !setBox() ) throw InvalidRegion();
} // PsfRegion::setRegion()

void PsfRegion::ellipseTrans(float rot, float major, float minor, float & Xmax, float & Ymax) {
	float cosRot = cos(rot/180*M_PI), sinRot = sin(rot/180*M_PI);
	float thetaX, thetaY;

	if (cosRot == 0)
		thetaX = M_PI / 2.0, thetaY = 0;
	else if (sinRot == 0)
		thetaX = 0, thetaY = M_PI / 2.0;
	else 	
		thetaX = atan(-minor / major * sinRot / cosRot),
		thetaY = atan(minor / major * cosRot / sinRot);

	Xmax = fabs(major * cos(thetaX) * cosRot - minor * sin(thetaX) * sinRot);
	Ymax = fabs(major * cos(thetaY) * sinRot + minor * sin(thetaY) * cosRot);
} // PsfRegion::ellipseTrans

void PsfRegion::boxTrans(float rot, float x, float y, float & Xmax, float & Ymax){
	float sinRot = sin(rot/180*M_PI), cosRot = cos(rot/180*M_PI);
	float cosX = cosRot * x, sinX = sinRot * x, cosY = cosRot * y, sinY = sinRot * y;
	float x1 = fabs(cosX + sinY), x2 = fabs(cosX - sinY);
	float y1 = fabs(sinX + cosY), y2 = fabs(sinX - cosY);
	Xmax = x1 > x2 ? x1 : x2;
	Ymax = y1 > y2 ? y1 : y2;
} // PsfRegion::boxTrans()

bool PsfRegion::setBox() {
	if ( isEmpty() ) return false;

	FShape * shape = at(0)[0];
	float width = 0, height = 0, orgWidth = 0, orgHeight = 0, rot = 0;
	switch ( shape->type() ) {
		case CIRCLE:
		case ANNULUS:
                       	height = width = shape->radii()[shape->type() == ANNULUS];
 			_RegionType = CIRCLE;
			break;
		case ELLIPSE:
		case ELLIPTANNULUS:
			orgWidth = shape->radii()[(shape->type() == ELLIPTANNULUS) * 2];
			orgHeight  = shape->radii()[(shape->type() == ELLIPTANNULUS) * 2 + 1];
			cout << orgWidth << "+" << orgHeight << endl;
			rot = shape->rotAngles()[shape->type() == ELLIPTANNULUS];
			ellipseTrans(rot, orgWidth, orgHeight, width, height);
			_RegionType = ELLIPSE;
			break;
		case BOX:
			width = shape->radii()[0] / 2.0;
			height = shape->radii()[1] / 2.0;
			_RegionType = BOX;
			break;
		case ROTBOX:
			orgWidth = shape->radii()[0] / 2.0;
			orgHeight = shape->radii()[1] / 2.0;
			rot = shape->rotAngles()[0];
			boxTrans(rot, orgWidth, orgHeight, width, height);
			_RegionType = BOX;
			break;
		default: return false;
	}

	_CoverBox = new FBox(shape->refPoint(), width*2.0, height*2.0);

	return true;
} // PsfRegion::setBox()

std::ostream & operator<< (ostream & os, const PsfRegion & reg) {
	if ( reg.isEmpty() ) 
		return (os << " Empty region!! " << endl);

	os << "Expression: " << reg.getExpression() << endl;
	
	for (unsigned int i = 0; i < reg.size(); i++) {
		os << " Component " << i << " has the following shapes: \n";
		for (unsigned int j = 0; j < reg[i].size(); j++) {
			os << "\t" << reg[i][j]->name() << ": \tcentre " << reg[i][j]->refPoint() << "\t";
			switch (reg[i][j]->type()) {
				case ShapeBasics::CIRCLE: {
					FCircle * regij = dynamic_cast<FCircle *> (reg[i][j]);
					os << "radius - " << regij->radius() << endl;
					break;
				}
				case ShapeBasics::ANNULUS: {
					FAnnulus * regij = dynamic_cast<FAnnulus *> (reg[i][j]);
					os << "radius - " << regij->innerCircle().radius() << " (inner) & " << regij->outerCircle().radius() << " (outer)" << endl;
					break;
				}
				case ShapeBasics::ELLIPSE: {
					FEllipse * regij = dynamic_cast<FEllipse *> (reg[i][j]);
					os << "major - " << regij->radiusMajor() << " minor - " << regij->radiusMinor() << " rotate - " << regij->rotAngle() << endl;
					break;
				}
				case ShapeBasics::ELLIPTANNULUS: {
					FElliptAnnulus * regij = dynamic_cast<FElliptAnnulus *> (reg[i][j]);
					os << "major - " << regij->innerEllipse().radiusMajor() << " minor - " << regij->innerEllipse().radiusMinor() << " rotate - " << regij->innerEllipse().rotAngle() << " (inner) & " << "major - " << regij->outerEllipse().radiusMajor() << " minor - " << regij->outerEllipse().radiusMinor() << " rotate - " << regij->outerEllipse().rotAngle() << " (outer)" << endl;
					break;
				}
				case ShapeBasics::BOX: {
					FBox * regij = dynamic_cast<FBox *> (reg[i][j]);
					os << "width - " << regij->xWidth() << " height - " << regij->yWidth() << endl;
					break;
				}
				case ShapeBasics::ROTBOX: {
					FRotBox * regij = dynamic_cast<FRotBox *> (reg[i][j]);
					os << "width - " << regij->xWidth() << " height - " << regij->yWidth() << " rotate - " << regij->rotAngle() << endl;
					break;
				}
				default: 	os << "Not supported shape!!!" << endl;
			}
		}
	}

	os << "Cover box: center " << reg.getCoverBox()->refPoint() << "width - " << reg.getCoverBox()->xWidth() << " height - " << reg.getCoverBox()->yWidth() << endl;

	os << "Type: ";
        switch (reg.getRegionType()) {
                case ShapeBasics::CIRCLE: os << " CIRCLE" << endl; break;
                case ShapeBasics::BOX: os << " BOX" << endl; break;
                case ShapeBasics::ELLIPSE: os << " ELLIPSE" << endl; break;
                default: os << " ERROR!!!!\n";
 	}

	return (os << endl << "CoorType: " << reg.getCoordType() << endl);
} // ostream & operator<< (ostream &, const PsfRegion &)
