//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//      PsfState.h:       --- header for PsfState class
//
//      DESCRIPTION:
//              An instance of class <PsfState> records the needed information 
//		for generating a specific PSF. It contains the following:
//                      o information for calServer setting
//                      o energy
//                      o PSF center in TELCOORD (theta & phi)
//                      o PSF size in pixels
//			o PSF file name
//
//      STRUCTURAL IDEA:
//              <PsfState> has a <CalState *> member which will be pointed to correct
//		<CalStateXXX> object according to the instrument value. Then calServer
//		settings will be set into this object for later calServer.setState().

#ifndef PSFSTATE_H
#define PSFSTATE_H

#include <string>
#include <vector>
#include <iostream>

#include "param.h"
#include "CalState.h"

using namespace std;

class PsfState {
public:
  PsfState() : _instrument(Xmm::EMOS1, "Instrument", "", "Telescope of Cal server", true, true, Xmm::INSTRUMENT_NONE, Xmm::OM),
               _state(NULL),
               _ev(1000., "Energy", "eV", "The PSF of this energy", true, true, 100., 15000.),
               _theta(0., "Theta", "arcsec", "off-axis angle", true, true, 0., 1800.),
               _phi(0., "Phi", "radian", "azimuth angle", true, true, -2*M_PI, 2*M_PI),
	       _rotate(0., "Theta", "radian", "rotation angle", true, true, -2*M_PI, 2*M_PI),
	       _xangle(0., "XAngle", "radian", "X transform angle", true, true, -2*M_PI, 2*M_PI),
               _yangle(0., "YAngle", "radian", "Y transform angle", true, true, -2*M_PI, 2*M_PI),
               _xsize(200, "XSize", "pixel", "PSF width", true, true, 1, 1024),
               _ysize(200, "YSize", "pixel", "PSF height", true, true, 1, 1024),
               _file("temp.fits", "PSF", "", "The output PSF file name", false, false, "", "") {};

    ~PsfState() throw() { try {if (_state) delete _state;} catch (...) {cout << "something wrong when deconstruct PsfState";} };

  CalState * state() const { return _state; };
  RangeStateVariable<Xmm::Instrument> * instrument() const { return &_instrument; };
  RangeStateVariable<double> * ev() const { return &_ev; };
  RangeStateVariable<double> * theta() const { return &_theta; };
  RangeStateVariable<double> * phi() const { return &_phi; };
  RangeStateVariable<double> * rotate() const { return &_rotate; };
  RangeStateVariable<float> * xangle() const { return &_xangle; };
  RangeStateVariable<float> * yangle() const { return &_yangle; };
  RangeStateVariable<int> * xsize() const { return &_xsize; };
  RangeStateVariable<int> * ysize() const { return &_ysize; };
  RangeStateVariable<string> * file() const { return &_file; };

  void setValueFromPara( int ev_p = 0 );

private:
  mutable RangeStateVariable<Xmm::Instrument> _instrument;
  mutable CalState * _state;
  mutable RangeStateVariable<double> _ev;
  mutable RangeStateVariable<double> _theta;
  mutable RangeStateVariable<double> _phi;
  mutable RangeStateVariable<double> _rotate;
  mutable RangeStateVariable<float> _xangle;
  mutable RangeStateVariable<float> _yangle;
  mutable RangeStateVariable<int> _xsize;
  mutable RangeStateVariable<int> _ysize;
  mutable RangeStateVariable<string> _file;
};

ostream &operator << (ostream &os, const PsfState &state);

#endif
