/*
 psfgen

   Authors:  Tao Song - University of Leicester
         
   Description: Generate PSF from ccf
 
   Comments: 
 
  $Id: module.cc,v 1.5 2000/10/02 08:30:13 gvacanti Exp $
*/

#include "psfgen.h"

#include <Tracer.h>
#include <errstr.h>

#include "PsfState.h"
#include "CalServer.h"
#include "DataSubspace.h"
#include "param.h"
#include "PsfRegion.h"
#include "ASCregion.h"
#include "PsfPosition.h"
#include "STime.h"

typedef double	real;
typedef bool	boolean;
typedef vector<double>	realList;

#define GETVALUE(name, type)    type name = type##Parameter(#name);
#define SETVALUE(name, value)          setParameter(#name, value);
#define SHOWSTR(name)	{GETVALUE(name, string); cout << #name << " = " << name << endl;}
#define SHOWINT(name)	{GETVALUE(name, int); cout << #name << " = " << name << endl;}
#define SHOWBOOL(name)	{GETVALUE(name, boolean); cout << #name << " = " << (name ? "yes" : "no") << endl;}
#define SHOWREAL(name)	{GETVALUE(name, real); cout << #name << " = " << name << endl;}
#define SHOWTIME(name)	{GETVALUE(name, real); cout << #name << " = " << STime(Xmm::missionReferenceTime + name) << endl;}
#define SHOWLIST(name)	{GETVALUE(name, realList); cout << #name << " = [ "; for (realList::iterator i = name.begin(); i < name.end(); ++i) cout << *i << ", "; cout << "]\n";}

void ShowParameters()
{
    
	SHOWSTR(instrument);
	SHOWSTR(image);
	SHOWBOOL(useodf);
	SHOWSTR(region);
	SHOWINT(ccdnr);
	SHOWSTR(node);
	SHOWSTR(cammode);
	SHOWSTR(filter);
	SHOWREAL(ccdtemp);
	SHOWREAL(camtemp);
	SHOWBOOL(temptrack);
	SHOWTIME(starttime);
	SHOWREAL(duration);
	SHOWSTR(level);
	SHOWINT(parameter);
	SHOWBOOL(random);
	SHOWINT(binning);
	SHOWSTR(ccf);
	SHOWBOOL(ccflog);
	SHOWLIST(energy);
	SHOWLIST(weight);
	SHOWREAL(theta);
	SHOWREAL(phi);
	SHOWREAL(rotate);
	SHOWINT(xsize);
	SHOWINT(ysize);
	SHOWSTR(output);
}

real32 * TransformPSF(Image<float> &image, Array *array, float PA, float xAngle, float yAngle) {
	// dimensions of Image and data pointer
	unsigned long rowImage = image.xExtent(), colImage = image.yExtent();
	float *imageData = image.data();

	// dimensions of Array and data pointer
	unsigned long rowArray = array->size()[0], colArray = array->size()[1];
	float *arrayData = array->data()->real32Data();
	unsigned long xImage = colImage>>1, yImage = rowImage>>1;	// center of the Image
	unsigned long xArray = colArray>>1, yArray = rowArray>>1;	// center of the Array
	//float temp=M_PI*1.5;

	// parameters for transform matrix
	//PA = 2 * (PA - temp);
	
	double a11 = cos(xAngle) * cos(PA);
	double a12 = cos(yAngle) * sin(PA);
	double a21 = cos(xAngle) * sin(PA);
	double a22 = cos(yAngle) * cos(PA);
	unsigned long x, y, n;
	float i, j, px, py, pxy;
	unsigned long sx01, sx23, sy03, sy12;

	--rowImage, --colImage;
	for (y = 0, n = 0; y < rowArray; ++y)
		for (x = 0; x < colArray; ++x) {
			// coordinates transform
			px = x, py = y;
			px -= xArray, py -= yArray;
			i = a11 * px - a12 * py + xImage;
			j = a21 * px + a22 * py + yImage;

			// find the surrounding 4 points' coordinates in imageData
			if (i < 0) {i = 0, sx01 = 0, sx23 = 1;}
			else if (i > colImage) {i = colImage, sx01 = colImage - 1, sx23 = colImage;}
			else sx01 = static_cast<unsigned long>(i), sx23 = sx01 + 1;

			if (j < 0) {j = 0, sy03 = 0, sy12 = 1;}
			else if (j > rowImage) {j = rowImage, sy03 = rowImage - 1, sy12 = rowImage;}
			else sy03 = static_cast<unsigned long>(j), sy12 = sy03 + 1;

			// linear interpolation
			// formula = P0 + px * v(P3) + py * v(P1) + pxy * (v(P2) - v(P1) - v(P3))
			//         = (1-px-py+pxy)*P0 + pxy*P2 + (py-pxy)*P1 + (px-pxy)*p3
			// where v(X) is the vector calculated from X-P0; pxy = px * py
			px = i - sx01, py = j - sy03, pxy = px * py;
			sy03 *= (colImage + 1), sy12 *= (colImage + 1);
			arrayData[n++] = (1 - px - py + pxy) * imageData[sx01 + sy03] + pxy * imageData[sx23 + sy12]
					+ (py - pxy) * imageData[sx01 + sy12] + (px - pxy) * imageData[sx23 + sy03];
//			pxy = 0;
//			arrayData[n++] = imageData[static_cast<unsigned long>(j+0.5)*(colImage+1) + static_cast<unsigned long>(i+0.5)];
		}

	return arrayData;
}

void psfgen ()
//int main()
{
PROC
   // get the image parameter to see whether an image is supplied
   string fitsfile = stringParameter("image");

   unsigned count = parameterCount("energy");
   char buff[50]; buff[49] = 0;
   Array *Normal = NULL;
   ArrayData *Normald = NULL;
   real32 *dNormal = NULL;
   
   PsfPosition psfPos;
   CoverFBox box;

   // if image parameter is given, read parameters from image dataset
   if (!fitsfile.empty()) { // image dataset supplied
      psfPos.setDataSet(fitsfile, stringParameter("region"));

      //overwrite instrument if not supplied in parameter
      GETVALUE(instrument, string);
      if (instrument.empty()) SETVALUE(instrument, psfPos.instrument());

      //overwrite starttime and duration if not supplied in parameter
      GETVALUE(duration, real);
      if (duration == 0) {
         STime start(psfPos.obsStart()), end(psfPos.obsEnd());
         SETVALUE(starttime, (start - Xmm::missionReferenceTime));
         SETVALUE(duration, (end - start));
      }

      //The following parameters can be detected by region, but because psfPos is initialised with region expression already, 
      //the parameters of psfPos should correspond to the region if there is a region string. So these parameters can be 
      //overwritten by the values of psfPos.
      box = psfPos.coverBox();
//cout << box; cout << psfPos;
      
      GETVALUE(ccdnr, int);
      if (ccdnr == 0) SETVALUE(ccdnr, psfPos.ccdnr());

//	cout << box.refPoint() << endl;
      GETVALUE(theta, real);
      if (theta < 0.0) SETVALUE(theta, box.refPoint().x());
      GETVALUE(phi, real);
      if (phi < 0.0) SETVALUE(phi, box.refPoint().y());
      GETVALUE(rotate, real);
      if (rotate == -1.0 ) SETVALUE(rotate, box.pa());
      GETVALUE(xangle, real);
      if (xangle < 0.0) SETVALUE(xangle, box.xAngle());
      GETVALUE(yangle, real);
      if (yangle < 0.0) SETVALUE(yangle, box.yAngle());
      
      GETVALUE(xsize, int);
      if (xsize == 0) //SETVALUE(xsize, static_cast<int>(box.xWidth() / 22.0 + 0.5));
         SETVALUE(xsize, static_cast<int>(box.xOrgDetSize() / 22.0 + 0.5));
      GETVALUE(ysize, int);
      if (ysize == 0) //SETVALUE(ysize, static_cast<int>(box.yWidth() / 22.0 + 0.5));
         SETVALUE(ysize, static_cast<int>(box.yOrgDetSize() /22.0 + 0.5));
   } else { // NO image dataset supplied
      //if no way to detect instrument, default it to "EMOS1"
      GETVALUE(instrument, string);
      if (instrument.empty()) SETVALUE(instrument, "EMOS1");
      
      //if no way to detect duration, default it to 3000.0
      GETVALUE(duration, real);
      if (duration == 0) SETVALUE(duration, 3000.0);

      //the following parameters can be detected by region
      GETVALUE(region, string);
         
      //if no way to detect ccd no, default it to 1
      GETVALUE(ccdnr, int);
      if (ccdnr == 0) SETVALUE(ccdnr, 1);

      //if no way to detect theta, phi, default it to 0
      GETVALUE(theta, real);
      if (theta < 0.0) SETVALUE(theta, 0.0);
      GETVALUE(phi, real);
      if (phi < 0.0) SETVALUE(phi, 0.0);
      GETVALUE(xangle, real);
      if (xangle < 0.0) SETVALUE(xangle, 0.0);
      GETVALUE(yangle, real);
      if (yangle < 0.0) SETVALUE(yangle, 0.0);

      //if no way to detect xsize, ysize, default it to 199
      GETVALUE(xsize, int);
      if (xsize == 0) SETVALUE(xsize, 199);
      GETVALUE(ysize, int);
      if (ysize == 0) SETVALUE(ysize, 199);
   }


   // Get all the default parameters
   PsfState state;
   state.setValueFromPara();
   string psffile = state.file()->get();

   // if no output, return the theta, phi, rotation, xsize, ysize
   if (psffile.empty()) {
      cout << "theta=" << state.theta()->get() << " phi=" << realParameter("phi") << " rotate=" << realParameter("rotate") << " xangle=" << realParameter("xangle") << " yangle=" << realParameter("yangle") << " xsize=" << state.xsize()->get() << " ysize=" << state.ysize()->get() << endl;

      return;
   }

   // Need to generate a PSF but without given energy
   if (!count) {
     std::cerr << "psfgen.cc: energy value missing!!" << endl;
     return;
   }

   if (state.state() == NULL) {
     std::cerr << "psfgen.cc: Cannot create a state!!" << endl;
     return;
   }

   DataSet *dataSet = dataSetServer->open(psffile, Dal::Create);

//   const XmmInstrument * oldInstrument = calServer.instrument();
   
   calServer.setInstrument(state.instrument()->get());
   calServer.pushState();
   calServer.atomServer()->restart();
   calServer.setState(state.state()[0]);

   string ccf = stringParameter("ccf");
   if (ccf.length()) {
	cout << "use user own ccf : " << ccf << endl;
	calServer.setCcf(ccf);
   }

   PsfDataServer *pds = 0; pds = calServer.getAtom(pds);
   std::vector<unsigned long> extents(2);
   extents[0] = state.ysize()->get();  //rows of expected PSF image
   extents[1] = state.xsize()->get();  //cols of expected PSF image

   for (unsigned i = 0; ; ) {
      Psf &psf = pds->psf(state.ev()->get(), state.theta()->get(), state.phi()->get());
      
      if (realParameter("rotate") >= 0) /*psf.setPosAngle(M_PI * 1.5 - state.rotate()->get())*/; 
      else cout << "rotate is not used.\n";

      // set the correct size for image windows
      
      float win_col, win_row;
      
      switch (state.state()->accuracyLevel()->get())
      {
        case LOW:
        case MEDIUM:
        case HIGH:	if (fitsfile.empty()) win_row = (extents[0] + 1) / 50.0, win_col = (extents[1] + 1) /50.0;
			else win_row = (box.yWidth() / 22.0 + 1) / 50.0, win_col = (box.xWidth() / 22.0 + 1) / 50.0;
			break;
        case EXTENDED: 
//        case ELLBETA:
			if (fitsfile.empty()) win_row = extents[0] / 55.0, win_col = extents[1] / 55.0;
			else win_row = box.yWidth() / 22.0 / 55.0, win_col = box.xWidth() / 22.0 / 55.0;
			break;
        default: std::cerr << "psfgen.cc: Unrecognised accuracy level!!" << endl; return;
      }

      Image<float> image = psf.image(SpatialWindow<double>(win_row, win_col));
      
      std::vector<unsigned long> realext(2);
      realext[0] = image.xExtent();   //rows of real PSF image
      realext[1] = image.yExtent();   //cols of real PSF image
   
      // because the output of ELLBETA level is odd and we can figure out what is the problem, we flush the output and return 
//      if (state.state()->accuracyLevel()->get() == ELLBETA) {
//        //unsigned long temp = extents[0];
//	//extents[0] = extents[1], extents[1] = temp;
//	dataSet << image;//	break;
//      }
	
      // when energy is a list more than noe value, create the primary image as the combination of all subsets
      if (Normal == NULL && count > 1) {
	Normal = dataSet->addArray("PRIMARY", Dal::Real32, extents);
	*(Normal->addAttribute("ENERGY", "energy value of the PSF")) = "Summation";
	Normald = Normal->data();
	dNormal = Normald->real32Data();
	unsigned nElements = Normal->elements();
	for (unsigned n=0; n<nElements; ++n) dNormal[n] = 0;
      }

      // when energy is a list has only one value, the primary image is the specified energy psf itself
      if (i==0 && count==1)
	sprintf(buff, "PRIMARY%c", 0);
      else
	sprintf(buff, "PSF_%f%c", realParameter("energy", i), 0);

      Array *a = dataSet->addArray(buff, Dal::Real32, extents);
      sprintf(buff, "%f%c", realParameter("energy", i), 0);
      *(a->addAttribute("ENERGY", "energy value of the PSF")) = buff;
      unsigned nElements = a->elements();
/*
      unsigned nImageElements = image.size();
      ArrayData *ad = a->data();
      real32 *d = ad->real32Data();

      // copy the pixels from image to the array
      float *imageData = image.data();

      for (unsigned n=0, y=(realext[0]-extents[0]+1)/2*realext[1]; y<nImageElements && n<nElements; y+=realext[1])
      	for (unsigned x=(realext[1]-extents[1]+1)/2, k = 0; x<realext[1] && k<extents[1]; x++, k++)
	  d[n++] = imageData[y+x]; 
*/
///*
      float pa = 0;
      if (realParameter("rotate") >= 0) pa = state.rotate()->get();
//      if (state.state()->accuracyLevel()->get() == ELLBETA) pa = -pa;
      real32 *d = TransformPSF(image, a, pa, state.xangle()->get(), state.yangle()->get());
//*/
      if (dNormal) {
	unsigned n = parameterCount("weight");
	float w = realParameter("weight", i<n ? i : n-1);
	for (unsigned n=0; n<nElements; ++n) dNormal[n] += d[n]*w;
      }
      
//      delete ad;

      if (++i >= count) break;
      
      state.setValueFromPara(i);
   } 
   
   if (Normald) delete Normald;

   dataSetServer->close(dataSet);

   if (booleanParameter("ccflog")) std::cerr << calServer.ccfAccessLog();

   calServer.popState();
   //calServer.setInstrument(oldInstrument->id());

ENDPROC
}
